# Copyright Seth Wenchel 2015

#' @param  .data A tbl or soomething that can be coerced into one
#' @param ... conditions that will be passed to dplyr::filter
#' @param false_fun A function or functional that will be applied to the data that doesn't pass the supplied filters (the scion)
#' @param false_name optional, the name of the object to which the scion will be assigned. If specified, false_env must also be specified
#' @param false_env option, the environment into which the scion will be assigned. If specified, false_name must also be specified
#' @return A tbl whose rows have passed the stated conditions
#' @details .data will be split into two chunks based on the conditions. The scion will be passed through false_fun and then either placed on
#' the package's internal stack or assigned as specified by false_name and false_env.
#' @examples
#' library(dplyr)
#' aframe <- data.frame(zed = runif(100))
#' set_to_zero <- . %>% mutate(zed = 0)
#' aframe %>% scion(zed >0.5, false_fun=set_to_zero) %>% mutate(zed=1) %>% graft
#'
#' @export
#' @title scion
#' @author Seth Wenchel
#' @importFrom dplyr filter
#' @importFrom dplyr setdiff
#' @importFrom magrittr %>%
scion <- function(.data, ..., false_fun, false_name, false_env){
  .data %>% dplyr::filter(...) -> tru
  falls <- dplyr::setdiff(.data, tru)
  if(!missing(false_fun))
    falls <- false_fun(falls)
  if(!missing(false_name)){
    if(!missing(false_env))
      assign(false_name, falls, envir = false_env)
    else
      stop("false_name specified but not false_env.  you must specify both or neither")
  }
  else
    if(!missing(false_env))
      stop("false_name specified but not false_env.  you must specify both or neither")
  .push(falls)
  return(tru)
}


#' @param .data A tbl or soomething that can be coerced into one
#' @param data2 A tbl or soomething that can be coerced into one
#' @param combine_fun optional, a function that will combbine two tbls such as full_join or bind_rows
#' @description Graft one dataset onto another
#' @details Graft requires two data objects.  The first must be provided by the user. The second can either be passed
#' in or automatically pulled off of the package's internal stack of scions. These will be combined accoring to the following rules in order:
#' * If either dataset has zero rows, the other dataset will be returned.
#' * If combine_fun is specifed, combine_fun(.data, data2) will be called
#' * If all column names match, a row bind will occur
#' * If at least some column names match, a full join will occur
#' * If both have the same number of rows a column bind will be performed
#' @return A single tbl object
#' @export
#' @author Seth Wenchel
#' @title Graft
#' @importFrom dplyr bind_rows
#' @importFrom dplyr bind_cols
#' @importFrom dplyr intersect
#' @importFrom dplyr full_join
graft <- function(.data, combine_fun, data2){
  if(missing(data2))
    data2 <- .pop()

  if(nrow(.data)==0)
    return(data2)
  if(nrow(data2)==0)
    return(.data)
  if(!missing(combine_fun))
    return(combine_fun(.data,data2))
  if(names(.data)==names(data2))
    return(dplyr::bind_rows(.data, data2))
  if(length(dplyr::intersect(names(.data),names(data2)))>0)
    return(dplyr::full_join(.data, data2))
  if(nrow(.data)==nrow(data2))
    return(dplyr::bind_cols(.data, data2))


  # same # columns & names <=> rbind
  # same # rows & no names match <=> cbind
  # at least one column names matches <=> full_join
}



# hark <- . %>% mutate(zed =0)
#
# aframe <- data.frame(zed = runif(100))
#
# aframe %>% scion(hark, zed >0.5) %>% mutate(zed=1) %>% graft
